/*
 * DBF/NTX reader.
 * Written by Boris Botstein.
 *
 * Source:
 * Rick Spence, "Clipper Programming Guide", Mikrotrend Books.
 */

#ifndef __DBF_OBJ
#define __DBF_OBJ

#include "dbfutls.hpp"

//#define USE_NT_CALLS

class ntx_file {
private:
	struct ntx_root {
		unsigned short signature;          // must be equal to 03.
		unsigned short version;            // index system version.
		long root;                         // first page offset.
		long unused;                       // first unused page offset.
		unsigned short item_size;          // distance between keys in page.
		unsigned short key_size;           // key size.
		unsigned short key_dec;            // for numeric keys.
		unsigned short max_item;           // max count of keys, holded in page.
		unsigned short half_page;          // min count of keys, holded in page.
		char key_expr[NTX_MAX_KEY_LENGTH]; // expression.
		unsigned short unique;             // unique flag.
	};

	struct ntx_item {
		long page;
		long rec_no;
		char key[NTX_MAX_KEY_LENGTH];
	};

	class ntx_page {
	private:
   	char* page;
		long parent;
		short entries;
		unsigned short item_size;
		mutable unsigned short index;

	public:
#ifdef USE_NT_CALLS
		ntx_page(HANDLE id, long _parent, const ntx_file::ntx_root& root)
			throw(runtime_error);
#else
		ntx_page(FILE* id, long _parent, const ntx_file::ntx_root& root)
			throw(runtime_error);
#endif

		~ntx_page() {
			if(page) delete[] page;
		}

		ntx_item* operator[](unsigned short i) const {
			if(entries > 0) {
				index = i;
				return (ntx_item*)(page + i * item_size);
			}
			return 0;
		}

		bool valid() { return entries > 0; }
		operator ntx_item*() const { return operator[](index); }
		ntx_item* _fastcall find(long offset) const throw(runtime_error);
		int _fastcall find(const char* key, int key_len) const;
		long get_parent() const { return parent; }
		short get_entries() const { return entries; }
		unsigned short get_index() const { return index; }
		bool on_last() const { return index == entries; }
	};

	typedef map< long, ntx_page*, less<long> > ntx_table;

	ntx_file(ntx_file&); // Performs link error on call.
	ntx_file& operator=(const ntx_file& rhs); // too.

#ifdef USE_NT_CALLS
	HANDLE ntx_handle;
#else
	FILE* ntx_handle;
#endif

	char vbuf[1024];

	char* name;

	ntx_root root;
	ntx_table image;

	ntx_page* page;
	long offset;

	char key[NTX_MAX_KEY_LENGTH + 1];

	void _fastcall set_key(const char* value = 0);
	int _fastcall compare_with(const char* value);

protected:
	void release();

	void get_root() throw(runtime_error);
	long get_first_page() const { return root.root; }

	void _fastcall get_disk_page(long _offset, long parent) throw(runtime_error);
	void _fastcall get_page(long _offset, long parent) throw(runtime_error);
	void _fastcall get_page(long _offset) throw(runtime_error);

	long get_next_up();
	long get_next_down();
	long get_prev_up();
	long get_prev_down();

public:
	ntx_file(const char* filename) throw(runtime_error);
	~ntx_file();

	void reset();

	long get_top();
	long get_bottom();
	long get_next();
	long get_prev();

	long find(const char* k, unsigned short k_len, int& compare);

	const char* get_key() const {
		if(page) if(page->valid()) return ((ntx_item*)(*page))->key;
		return 0;
	}

	unsigned short get_key_size() const { return root.key_size; }
};

class dbf_field {
private:
	friend class dbf_file;

	mutable char field[DBF_MAX_FIELD_LENGTH];
	char* ptr;
	int size, dec, offset;
	char type;

	dbf_field(char* buffer, const dbf_field_desc& desc);
	dbf_field();

	void initialize() const {
		::memcpy(field, ptr + offset, size);
		field[size] = 0;
	}

public:
	int get_size() const { return size; }

	operator DateTy() const;
	operator int() const;
	operator double() const;
	operator char*() const;
	operator bool() const;

	char* ansi() const;
};

class dbf_file {
protected:
	typedef map< int, dbf_field_desc*, less<int> > dbf_field_list;

	dbf_header header;
	dbf_field_list fields;
	int fields_no;

#ifdef USE_NT_CALLS
	HANDLE dbf_handle;
#else
	FILE* dbf_handle;
#endif

	char* buffer;
	long record;
	ntx_file* index;
	char* name;
	char* vbuf;

	dbf_file(dbf_file&); // Performs link error on call.
	dbf_file& operator=(const dbf_file& rhs); // too.

	void _fastcall initialize(const char* filename, size_t bufsize)
		throw(runtime_error);
	bool _fastcall get_record(long rec_no);

public:
	dbf_file(const char* filename, size_t bufsize = 0);
	dbf_file(const char* dbf, const char* ntx, size_t bufsize = 0);
	virtual ~dbf_file();

	void reset() {
		record = 0;
		if(buffer) ::memset(buffer, 0, header.rec_len + 1);
	}

	void reset_all() {
		reset();
		if(index) index->reset();
	}

	bool deleted() {
		if(buffer) return buffer[0] == '*';
		return false;
	}
	long current() { return record; }
	const char* get_key() const {
		if(index) return index->get_key();
		return 0;
	}
	const char* get_buffer() const { return buffer; }
	const char* get_name() const { return name; }

	bool get_top();
	bool get_bottom();
	bool get_next();
	bool get_prev();

	bool find(const char* k, unsigned short k_len);

	dbf_field get_field(int field_no);
	dbf_field get_field(const char* field_name);
};

class dbf_fly : public dbf_file {
public:
	dbf_fly(const char* filename, size_t bufsize = 0) :
		dbf_file(filename, bufsize) {}
	bool next();
};

#endif // __DBF_OBJ

