/*
 * DBF/NTX reader.
 * Written by Boris Botstein.
 */

#include <windows.h>
#include <stdio.h>
#include <share.h>

#pragma hdrstop
#include "dbfntx.hpp"

ntx_file::ntx_file(const char* filename) throw(runtime_error) :
	page(0), offset(0l), name(0) {
	name = new_str(filename);
#ifdef USE_NT_CALLS
	ntx_handle = ::CreateFile(filename, GENERIC_READ, FILE_SHARE_READ |
		FILE_SHARE_WRITE, 0, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);
	if(INVALID_HANDLE_VALUE == ntx_handle)
		throw runtime_error(_string("Can't open ") + _string(filename));    
#else
	ntx_handle = ::_fsopen(filename, "rb", SH_DENYNONE);
	if(!ntx_handle) throw runtime_error(_string("Can't open ") + _string(filename));
	::setvbuf(ntx_handle, vbuf, _IOFBF, sizeof(vbuf));
#endif
	reset();
}

ntx_file::~ntx_file() {
	if(name) delete[] name;
	release();
#ifdef USE_NT_CALLS
	::CloseHandle(ntx_handle);
#else
	::fclose(ntx_handle);
#endif
}

void _fastcall ntx_file::set_key(const char* value) {
	::memset(key, 0, sizeof(key));
	if(value) ::memcpy(key, value, get_key_size());
}

int _fastcall ntx_file::compare_with(const char* value) {
	return ::memcmp(key, value, get_key_size());
}

void ntx_file::get_root() throw(runtime_error) {

#ifdef USE_NT_CALLS
	if(::SetFilePointer(ntx_handle, offset, 0, FILE_BEGIN) == 0xFFFFFFFF)
#else
	if(::fseek(ntx_handle, offset, SEEK_SET))
#endif
		throw runtime_error(_string(name) + _string(" - lseek error (0)"));

#ifdef USE_NT_CALLS
	DWORD dwBytesRead;
	if(!::ReadFile(ntx_handle , &root, sizeof(ntx_root), &dwBytesRead, 0) ||
		dwBytesRead != sizeof(ntx_root))
#else
	if(::fread(&root, sizeof(ntx_root), 1, ntx_handle) != 1)
#endif
		throw runtime_error(_string(name) + _string(" - can't read header"));

#if 0
	::printf("NTX ROOT:\n");
	::printf("signature == %d\n", root.signature);
	::printf("version == %d\n", root.version);
	::printf("root == %ld\n", root.root);
	::printf("unused == %ld\n", root.unused);
	::printf("item size == %d\n", root.item_size);
	::printf("key size == %d\n", root.key_size);
	::printf("key dec == %d\n", root.key_dec);
	::printf("max item == %d\n", root.max_item);
	::printf("half page == %d\n", root.half_page);
	::printf("expression == %s\n", root.key_expr);
	::printf("unique == %d\n\n", root.unique);
#endif
	if((root.key_size > NTX_MAX_KEY_LENGTH) ||
		(root.item_size != (root.key_size + (sizeof(long) << 1))))
		throw runtime_error(_string(name) + _string(" - invalid root page"));
}

void _fastcall ntx_file::get_disk_page(long _offset, long parent)
	throw(runtime_error) {
	if(_offset % 1024)
		throw runtime_error(_string(name) + _string(" - invalid offset"));

#ifdef USE_NT_CALLS
	DWORD dwPos = SetFilePointer(ntx_handle, 0, NULL, FILE_CURRENT);
	if(dwPos != _offset)
		if(::SetFilePointer(ntx_handle, _offset, 0, FILE_BEGIN) == 0xFFFFFFFF)
#else
	if(::ftell(ntx_handle) != _offset)
		if(::fseek(ntx_handle, _offset, SEEK_SET))
#endif
			throw runtime_error(_string(name) + _string(" - lseek error (1)"));

	try {
		page = new ntx_page(ntx_handle, parent, root);
		offset = _offset;
		image.insert(ntx_table::value_type(_offset, page));
	}
	catch(exception& e) {
		throw runtime_error(_string(name) + _string(e.what()));
	}
}

void ntx_file::release() {
	map_release(image.begin(), image.end());
	image.erase(image.begin(), image.end());
	page = 0;
	offset = 0l;
}

void ntx_file::reset() {
	release();
	get_root();
	get_disk_page(get_first_page(), 0l);
	if(page->valid()) set_key((*page)[0]->key);
	else set_key();
}

void _fastcall ntx_file::get_page(long _offset, long parent)
	throw(runtime_error) {
	ntx_table::const_iterator i = image.find(_offset);
	if(i != image.end()) {
		offset = _offset;
		page = (*i).second;
	}
	else get_disk_page(_offset, parent);
}

void _fastcall ntx_file::get_page(long _offset) throw(runtime_error) {
	ntx_table::const_iterator i = image.find(_offset);
	if(i != image.end()) {
		offset = _offset;
		page = (*i).second;
	}
	else throw runtime_error(_string(name) + _string(" - can't find page"));
}

long ntx_file::get_top() {
	long current = 0l, next = get_first_page(), rec_no = 0l;
	ntx_item* item;
	while(next) {
		get_page(next, current);
		if(!page->valid()) return 0l;
		item = (*page)[0];
		current = next;
		next = item->page;
		rec_no = item->rec_no;
	}
	if(compare_with(item->key) < 0)
		throw runtime_error(_string(name) + _string(" - invalid order (top)"));
	set_key(item->key);
	return rec_no;
}

long ntx_file::get_bottom() {
	long current = 0l, next = get_first_page(), rec_no = 0l;
	unsigned short entries;
	ntx_item* item;
	while(next) {
		get_page(next, current);
		if(!page->valid()) return 0l;
		entries = page->get_entries();
		item = (*page)[entries];
		if(!item->page) {
			entries--;
			item = (*page)[entries];
		}
		current = next;
		next = item->page;
		rec_no = item->rec_no;
	}
	if(compare_with(item->key) > 0)
		throw runtime_error(_string(name) + _string(" - invalid order (bottom)"));
	set_key(item->key);
	return rec_no;
}

long ntx_file::get_next_up() {
	long rec_no = 0l;
	if(page) {
		if(!page->valid()) return 0l;
		long parent, current;
		ntx_item* item;
		do {
			current = offset;
			parent = page->get_parent();
			if(!parent) return 0l;
			get_page(parent);
			if(!page->valid()) return 0l;
			item = page->find(current);
		} while(page->get_index() == page->get_entries());
		rec_no = item->rec_no;
		if(compare_with(item->key) > 0)
			throw runtime_error(_string(name) + _string(" - invalid order (next_up)"));
		set_key(item->key);
	}
	return rec_no;
}

long ntx_file::get_next_down() {
	long rec_no = 0l, next, current;
	ntx_item* item;
	if(page) {
		if(!page->valid()) return 0l;
		item = *page;
		current = offset;
		next = item->page;
		while(next) {
			get_page(next, current);
			if(!page->valid()) return 0l;
			item = (*page)[0];
			rec_no = item->rec_no;
			current = next;
			next = item->page;
		}
		if(compare_with(item->key) > 0)
			throw runtime_error(_string(name) +
				_string(" - invalid order (next_down)"));
		set_key(item->key);
	}
	return rec_no;
}

long ntx_file::get_next() {
	if(page) {
		if(!page->valid()) return 0l;
		unsigned short item_no = page->get_index();
		if(item_no < page->get_entries()) item_no++;
		ntx_item* item = (*page)[item_no];
		if(item_no < page->get_entries()) {
			if(!item->page) {
				if(compare_with(item->key) > 0)
					throw runtime_error(_string(name) +
						_string(" - invalid order (next)"));
				set_key(item->key);
				return item->rec_no;
			}
			else return get_next_down();
		}
		else if(item->page) return get_next_down();
		else return get_next_up();
	}
	return 0l;
}

long ntx_file::get_prev_up() {
	long rec_no = 0l, parent;
	unsigned short item_no;
	ntx_item* item;
	if(page) {
		if(!page->valid()) return 0l;
		item_no = page->get_index();
		while(!item_no) {
			parent = page->get_parent();
			if(!parent) return 0l;
			get_page(parent);
			if(!page->valid()) return 0l;
			item_no = page->get_index();
		}
		item_no--;
		item = (*page)[item_no];
		rec_no = item->rec_no;
		if(compare_with(item->key) < 0)
			throw runtime_error(_string(name) +
				_string(" - invalid order (prev_up)"));
		set_key(item->key);
	}
	return rec_no;
}

long ntx_file::get_prev_down() {
	long rec_no = 0l, next, current;
	ntx_item* item;
	if(page) {
		if(!page->valid()) return 0l;
		item = *page;
		current = offset;
		next = item->page;
		while(next) {
			get_page(next, current);
			if(!page->valid()) return 0l;
			item = (*page)[page->get_entries()];
			current = next;
			next = item->page;
		}
		unsigned short item_no = page->get_entries();
		item_no--;
		item = (*page)[item_no];
		rec_no = item->rec_no;
		if(compare_with(item->key) < 0)
			throw runtime_error(_string(name) +
				_string(" - invalid order (prev_down)"));
		set_key(item->key);
	}
	return rec_no;
}

long ntx_file::get_prev() {
	if(page) {
		if(!page->valid()) return 0l;
		unsigned short item_no = page->get_index();
		ntx_item* item = (*page)[item_no];
		if(item->page) return get_prev_down();
		else if(item_no) {
			item_no--;
			item = (*page)[item_no];
			if(compare_with(item->key) < 0)
				throw runtime_error(_string(name) +
					_string(" - invalid order (prev)"));
			set_key(item->key);
			return item->rec_no;
		}
		else return get_prev_up();
	}
	return 0l;
}

long ntx_file::find(const char* k, unsigned short k_len, int& compare) {
	long current = 0l, next = get_first_page(), rec_no = 0l;
	ntx_item* item;
	int cmp;
	if(k_len > get_key_size()) k_len = get_key_size();
	bool now_ok = false;
	bool go_up = false;
	while(next) {
		if(go_up) get_page(next);
		else get_page(next, current);
		if(!page->valid()) return -1l;
		current = next;
		cmp = sign(page->find(k, k_len));
		item = *page;
		if(cmp == 0) {
			if(!go_up) {
				next = item->page;
				now_ok = true;
			}
			else next = 0l;
		}
		else {
			if(!go_up) go_up = page->on_last() ? false : now_ok;
			next = go_up ? page->get_parent() : item->page;
		}
		rec_no = item->rec_no;
	}
	set_key(item->key);
	compare = cmp;
	return rec_no;
}

