/*
 * DBF/NTX reader.
 * Written by Boris Botstein.
 */

#include <windows.h> 
#include <stdio.h>
#include <share.h>

#pragma hdrstop
#include "dbfntx.hpp"

dbf_file::dbf_file(const char* filename, size_t bufsize) :
	index(0), buffer(0), record(0l), name(0), dbf_handle(0), vbuf(0) {
	initialize(filename, bufsize);
	get_top();
}

dbf_file::dbf_file(const char* dbf, const char* ntx, size_t bufsize) :
	index(0), buffer(0), record(0l), name(0), dbf_handle(0), vbuf(0) {
	initialize(dbf, bufsize);
	index = new ntx_file(ntx);
	get_top();
}

dbf_file::~dbf_file() {
#ifdef USE_NT_CALLS
	::CloseHandle(dbf_handle);
#else
	if(dbf_handle) ::fclose(dbf_handle);
#endif
	if(buffer) delete[] buffer;
	if(name) delete[] name;
	if(vbuf) delete[] vbuf;
	if(index) delete index;
	map_release(fields.begin(), fields.end());
	fields.erase(fields.begin(), fields.end());
}

#ifdef USE_NT_CALLS
void _fastcall dbf_file::initialize(const char* filename, size_t /*bufsize*/)
#else
void _fastcall dbf_file::initialize(const char* filename, size_t bufsize)
#endif
	throw(runtime_error) {

#ifdef USE_NT_CALLS
	dbf_handle = ::CreateFile(filename, GENERIC_READ, FILE_SHARE_READ |
		FILE_SHARE_WRITE, 0, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);
	if(INVALID_HANDLE_VALUE == dbf_handle)
#else
	dbf_handle = ::_fsopen(filename, "rb", SH_DENYNONE);
	if(!dbf_handle)
#endif
		throw runtime_error(_string("Can't open ") + _string(filename));

#ifdef USE_NT_CALLS
	DWORD dwBytesRead;
	if(!::ReadFile(dbf_handle , &header, sizeof(dbf_header), &dwBytesRead, 0) ||
		dwBytesRead != sizeof(dbf_header))
#else
	if(::fread(&header, sizeof(dbf_header), 1, dbf_handle) != 1)
#endif
		throw runtime_error(_string(filename) +
			_string(" - can't read header"));

	fields_no = (header.length / (sizeof(dbf_field_desc) - sizeof(int))) - 1;
	for(int i = 0, offset = 1; i < fields_no; i++) {
		dbf_field_desc* field = new dbf_field_desc;

#ifdef USE_NT_CALLS
		DWORD dwBytesRead;
		if(!::ReadFile(dbf_handle , field, sizeof(dbf_field_desc) - sizeof(int),
			&dwBytesRead, 0) || dwBytesRead !=
			sizeof(dbf_field_desc) - sizeof(int)) {
#else
		if(::fread(field, sizeof(dbf_field_desc) - sizeof(int), 1, dbf_handle) != 1) {
#endif

			delete field;
			throw runtime_error(_string(filename) +
				_string(" - can't read field"));
		}
		field->offset = offset;
		offset += field->length;
		fields.insert(dbf_field_list::value_type(i, field));

#if 0
		::printf("name == %s\n", field->name);
		::printf("type == %c\n", field->type);
		::printf("length == %d\n", (int)field->length);
		::printf("dec == %d\n", (int)field->dec);
		::printf("offset == %d\n\n", field->offset);
#endif

	}
	buffer = new char[header.rec_len + 1];
	name = new_str(filename);

#ifndef USE_NT_CALLS
	if(bufsize) {
		vbuf = new char[bufsize];
		::setvbuf(dbf_handle, vbuf, _IOFBF, bufsize);
	}
#endif

}

bool _fastcall dbf_file::get_record(long rec_no) {
	if(rec_no < 1) {
		reset();
		return false;
	}
	long offset = header.length + (header.rec_len * (rec_no - 1));

#ifdef USE_NT_CALLS
	DWORD dwPos = SetFilePointer(dbf_handle, 0, NULL, FILE_CURRENT);
	if(dwPos != offset)
		if(::SetFilePointer(dbf_handle, offset, 0, FILE_BEGIN) == 0xFFFFFFFF) {
#else
	if(::ftell(dbf_handle) != offset)
		if(::fseek(dbf_handle, offset, SEEK_SET)) {
#endif

			reset();
			return false;
		}

#ifdef USE_NT_CALLS
	DWORD dwBytesRead;
	if(!::ReadFile(dbf_handle , buffer, header.rec_len, &dwBytesRead, 0) ||
		dwBytesRead != header.rec_len) {
#else
	if(::fread(buffer, header.rec_len, 1, dbf_handle) != 1) {
#endif

		reset();
		return false;
	}
	record = rec_no;
	return true;
}

bool dbf_file::get_top() {
	bool result;
	if(index) result = get_record(index->get_top());
	else result = get_record(1);
	while(result && deleted()) result = get_next();
	return result;
}

bool dbf_file::get_bottom() {
	bool result;
	if(!index) {
		result = get_record(header.rec_no + 1);
		if(!result) result = get_record(header.rec_no);
	}
	else result = get_record(index->get_bottom());
	while(result && deleted()) result = get_prev();
	return result;
}

bool dbf_file::get_next() {
	bool result = true;
	buffer[0] = '*';
	while(result && deleted()) {
		if(index) result = get_record(index->get_next());
		else result = get_record(record + 1);
	}
	return result;
}

bool dbf_file::get_prev() {
	bool result = true;
	buffer[0] = '*';
	while(result && deleted()) {
		if(index) result = get_record(index->get_prev());
		else result = get_record(record - 1);
	}
	return result;
}

// index->find compares:
//  0: search == key
// -1: search < key
//  1: search > key
bool dbf_file::find(const char* k, unsigned short k_len) {
	if(!k || !index) return false;
	int compare;
	if(k_len > index->get_key_size()) k_len = index->get_key_size();
	bool result = get_record(index->find(k, k_len, compare));
	while(result && deleted()) result = get_next();
	if(result) result = ::memcmp(k, index->get_key(), k_len) == 0;
	return result;
}

bool dbf_fly::next() {
	do {

#ifdef USE_NT_CALLS
		DWORD dwBytesRead;
		if(!::ReadFile(dbf_handle , buffer, header.rec_len, &dwBytesRead, 0) ||
			dwBytesRead != header.rec_len) {
#else
		if(::fread(buffer, header.rec_len, 1, dbf_handle) != 1) {
#endif

			reset();
			return false;
		}
		record++;
	}
	while(deleted());
	return true;
}

